@@ -7,10 +7,10 @@
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
- * This library is distributed in the hope that it will be useful, 
- * but WITHOUT ANY WARRANTY; without even the implied warranty of 
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details. 
+ * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
@@ -18,12 +18,20 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "config.h"
+#include <config.h>
+#include "exif-mnote-data.h"
+
 #include "exif-data.h"
 #include "exif-ifd.h"
+#include "exif-mnote-data-priv.h"
 #include "exif-utils.h"
 #include "exif-loader.h"
 #include "jpeg-marker.h"
+#include <libexif/exif-log.h>
+
+#include <libexif/olympus/exif-mnote-data-olympus.h>
+#include <libexif/canon/exif-mnote-data-canon.h>
+#include <libexif/pentax/exif-mnote-data-pentax.h>
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -32,7 +40,9 @@
 #undef MAX
 #define MAX(a, b)  (((a) > (b)) ? (a) : (b))
 
-//#define DEBUG
+#ifdef __WATCOMC__
+#      define strncasecmp strnicmp
+#endif
 
 static const unsigned char ExifHeader[] = {0x45, 0x78, 0x69, 0x66, 0x00, 0x00};
 
@@ -40,29 +50,67 @@ struct _ExifDataPrivate
 {
 	ExifByteOrder order;
 
+	ExifMnoteData *md;
+
+	ExifLog *log;
+	ExifMem *mem;
+
 	unsigned int ref_count;
+
+	/* Temporarily used while loading data */
+	unsigned int offset_mnote;
 };
 
+static void *
+exif_data_alloc (ExifData *data, unsigned int i)
+{
+	void *d;
+
+	if (!data || !i) return NULL;
+
+	d = exif_mem_alloc (data->priv->mem, i);
+	if (d) return d;
+
+	EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", i);
+	return NULL;
+}
+
+ExifMnoteData *
+exif_data_get_mnote_data (ExifData *d)
+{
+	return (d && d->priv) ? d->priv->md : NULL;
+}
+
 ExifData *
 exif_data_new (void)
+{
+	ExifMem *mem = exif_mem_new_default ();
+	ExifData *d = exif_data_new_mem (mem);
+
+	exif_mem_unref (mem);
+
+	return d;
+}
+
+ExifData *
+exif_data_new_mem (ExifMem *mem)
 {
 	ExifData *data;
 	unsigned int i;
 
-	data = malloc (sizeof (ExifData));
-	if (!data)
-		return (NULL);
-	memset (data, 0, sizeof (ExifData));
-	data->priv = malloc (sizeof (ExifDataPrivate));
-	if (!data->priv) {
-		free (data);
-		return (NULL);
-	}
-	memset (data->priv, 0, sizeof (ExifDataPrivate));
+	if (!mem) return NULL;
+
+	data = exif_mem_alloc (mem, sizeof (ExifData));
+	if (!data) return (NULL);
+	data->priv = exif_mem_alloc (mem, sizeof (ExifDataPrivate));
+	if (!data->priv) { exif_mem_free (mem, data); return (NULL); }
 	data->priv->ref_count = 1;
 
+	data->priv->mem = mem;
+	exif_mem_ref (mem);
+
 	for (i = 0; i < EXIF_IFD_COUNT; i++) {
-		data->ifd[i] = exif_content_new ();
+		data->ifd[i] = exif_content_new_mem (data->priv->mem);
 		if (!data->ifd[i]) {
 			exif_data_free (data);
 			return (NULL);
@@ -94,6 +142,10 @@ exif_data_load_data_entry (ExifData *data, ExifEntry *entry,
 	entry->format     = exif_get_short (d + offset + 2, data->priv->order);
 	entry->components = exif_get_long  (d + offset + 4, data->priv->order);
 
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Loading entry 0x%x ('%s')...", entry->tag,
+		  exif_tag_get_name (entry->tag));
+
 	/*
 	 * Size? If bigger than 4 bytes, the actual data is not
 	 * in the entry but somewhere else (offset).
@@ -110,47 +162,81 @@ exif_data_load_data_entry (ExifData *data, ExifEntry *entry,
 	if (size < doff + s)
 		return;
 
-	entry->data = malloc (sizeof (char) * s);
-	if (!entry->data)
-		return;
-	entry->size = s;
-	memcpy (entry->data, d + doff, s);
+	entry->data = exif_data_alloc (data, s);
+	if (entry->data) {
+		entry->size = s;
+		memcpy (entry->data, d + doff, s);
+	}
+
+#ifndef EXIF_DONT_CHANGE_MAKER_NOTE
+	/* If this is the MakerNote, remember the offset */
+	if (entry->tag == EXIF_TAG_MAKER_NOTE) {
+		if (entry->size > 6) exif_log (data->priv->log,
+			  EXIF_LOG_CODE_DEBUG, "ExifData",
+		          "MakerNote found (%02x %02x %02x %02x "
+			  "%02x %02x %02x...).",
+			  entry->data[0], entry->data[1], entry->data[2],
+			  entry->data[3], entry->data[4], entry->data[5],
+			  entry->data[6]);
+		data->priv->offset_mnote = doff;
+	}
+#endif
+
+	exif_entry_fix (entry);
 }
 
 static void
-exif_data_save_data_entry (ExifData *data, ExifEntry *entry,
+exif_data_save_data_entry (ExifData *data, ExifEntry *e,
 			   unsigned char **d, unsigned int *ds,
 			   unsigned int offset)
 {
 	unsigned int doff, s;
 
+	if (!data || !data->priv) return;
+
 	/*
 	 * Each entry is 12 bytes long. The memory for the entry has
 	 * already been allocated.
 	 */
 	exif_set_short (*d + 6 + offset + 0,
-			data->priv->order, entry->tag);
+			data->priv->order, (ExifShort) e->tag);
 	exif_set_short (*d + 6 + offset + 2,
-			data->priv->order, entry->format);
+			data->priv->order, (ExifShort) e->format);
+
+	/* If this is the maker note tag, update it. */
+	if ((e->tag == EXIF_TAG_MAKER_NOTE) && data->priv->md) {
+		exif_mem_free (data->priv->mem, e->data);
+		e->data = NULL;
+		e->size = 0;
+		exif_mnote_data_set_offset (data->priv->md, *ds - 6);
+		exif_mnote_data_save (data->priv->md, &e->data, &e->size);
+		e->components = e->size;
+	}
+
 	exif_set_long  (*d + 6 + offset + 4,
-			data->priv->order, entry->components);
+			data->priv->order, e->components);
 
 	/*
 	 * Size? If bigger than 4 bytes, the actual data is not in
 	 * the entry but somewhere else.
 	 */
-	s = exif_format_get_size (entry->format) * entry->components;
-	if (!s)
-		return;
+	s = exif_format_get_size (e->format) * e->components;
 	if (s > 4) {
-		*ds += entry->size;
-		*d = realloc (*d, sizeof (char) * *ds);
-		doff = *ds - 6 - entry->size;
+		doff = *ds - 6;
+		*ds += s;
+		*d = exif_mem_realloc (data->priv->mem, *d, *ds);
+		if (!*d) {
+			EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", *ds);
+		  	return;
+		}
 		exif_set_long (*d + 6 + offset + 8,
 			       data->priv->order, doff);
 	} else
 		doff = offset + 8;
-	memcpy (*d + 6 + doff, entry->data, entry->size);
+
+	/* Write the data. Fill unneeded bytes with 0. */
+	memcpy (*d + 6 + doff, e->data, s);
+	if (s < 4) memset (*d + 6 + doff + s, 0, (4 - s));
 }
 
 static void
@@ -158,16 +244,15 @@ exif_data_load_data_thumbnail (ExifData *data, const unsigned char *d,
 			       unsigned int ds, ExifLong offset, ExifLong size)
 {
 	if (ds < offset + size) {
-#ifdef DEBUG
-		printf ("Bogus thumbnail offset and size: %i < %i + %i.\n",
-			(int) ds, (int) offset, (int) size);
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "Bogus thumbnail offset and size: %i < %i + %i.",
+			  (int) ds, (int) offset, (int) size);
 		return;
 	}
-	if (data->data)
-		free (data->data);
+	if (data->data) exif_mem_free (data->priv->mem, data->data);
 	data->size = size;
-	data->data = malloc (sizeof (char) * data->size);
+	data->data = exif_data_alloc (data, data->size);
+	if (!data->data) return;
 	memcpy (data->data, d + offset, data->size);
 }
 
@@ -182,19 +267,21 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 	unsigned int i;
 	ExifTag tag;
 
+	if (!data || !data->priv) return;
+
 	/* Read the number of entries */
+	if (offset >= ds - 1) return;
 	n = exif_get_short (d + offset, data->priv->order);
-#ifdef DEBUG
-	printf ("Loading %i entries...\n", n);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+	          "Loading %i entries...", n);
 	offset += 2;
+
+	/* Check if we have enough data. */
+	if (offset + 12 * n > ds) n = (ds - offset) / 12;
+
 	for (i = 0; i < n; i++) {
 
 		tag = exif_get_short (d + offset + 12 * i, data->priv->order);
-#ifdef DEBUG
-		printf ("Loading entry '%s' (%i of %i)...\n",
-			exif_tag_get_name (tag), i + 1, n);
-#endif
 		switch (tag) {
 		case EXIF_TAG_EXIF_IFD_POINTER:
 		case EXIF_TAG_GPS_INFO_IFD_POINTER:
@@ -217,9 +304,6 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 					data->ifd[EXIF_IFD_INTEROPERABILITY], d, ds, o);
 				break;
 			case EXIF_TAG_JPEG_INTERCHANGE_FORMAT:
-#ifdef DEBUG
-				printf ("Thumbnail at %i.\n", (int) o);
-#endif
 				thumbnail_offset = o;
 				if (thumbnail_offset && thumbnail_length)
 					exif_data_load_data_thumbnail (data, d,
@@ -227,9 +311,6 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 						thumbnail_length);
 				break;
 			case EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH:
-#ifdef DEBUG
-				printf ("Thumbnail size: %i.\n", (int) o);
-#endif
 				thumbnail_length = o;
 				if (thumbnail_offset && thumbnail_length)
 					exif_data_load_data_thumbnail (data, d,
@@ -241,7 +322,13 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 			}
 			break;
 		default:
-			entry = exif_entry_new ();
+
+			/*
+			 * If we don't know the tag, changes are high
+			 * that the EXIF data does not follow the standard.
+			 */
+			if (!exif_tag_get_name (tag)) return;
+			entry = exif_entry_new_mem (data->priv->mem);
 			exif_content_add_entry (ifd, entry);
 			exif_data_load_data_entry (data, entry, d, ds,
 						   offset + 12 * i);
@@ -259,8 +346,7 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 	unsigned int j, n_ptr = 0, n_thumb = 0;
 	ExifIfd i;
 
-	if (!data || !ifd || !d || !ds)
-		return;
+	if (!data || !data->priv || !ifd || !d || !ds) return;
 
 	for (i = 0; i < EXIF_IFD_COUNT; i++)
 		if (ifd == data->ifd[i])
@@ -303,19 +389,21 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 	 * and the number of entries.
 	 */
 	*ds += (2 + (ifd->count + n_ptr + n_thumb) * 12 + 4);
-	*d = realloc (*d, sizeof (char) * *ds);
+	*d = exif_mem_realloc (data->priv->mem, *d, *ds);
+	if (!*d) {
+		EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", *ds);
+	  	return;
+	}
 
 	/* Save the number of entries */
 	exif_set_short (*d + 6 + offset, data->priv->order,
-			ifd->count + n_ptr + n_thumb);
+			(ExifShort) (ifd->count + n_ptr + n_thumb));
 	offset += 2;
 
-#ifdef DEBUG
-	printf ("Saving %i entries (IFD '%s', offset: %i)...\n",
-		ifd->count, exif_ifd_get_name (i), offset);
-#endif
-
 	/* Save each entry */
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saving %i entries (IFD '%s', offset: %i)...",
+		  ifd->count, exif_ifd_get_name (i), offset);
 	for (j = 0; j < ifd->count; j++)
 		exif_data_save_data_entry (data, ifd->entries[j],
 				d, ds, offset + 12 * j);
@@ -402,14 +490,14 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 			exif_set_long  (*d + 6 + offset + 8, data->priv->order,
 					*ds - 6);
 			*ds += data->size;
-			*d = realloc (*d, sizeof (char) * *ds);
+			*d = exif_mem_realloc (data->priv->mem, *d, *ds);
+			if (!*d) {
+				EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData",
+						    *ds);
+			  	return;
+			}
 			memcpy (*d + *ds - data->size, data->data, data->size);
 			offset += 12;
-#ifdef DEBUG
-			printf ("Wrote %i bytes of thumbnail data at offset "
-				"%i.\n", data->size, *ds - data->size);
-			printf ("We currently have %i bytes EXIF data.\n", *ds);
-#endif
 
 			/* EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH */
 			exif_set_short (*d + 6 + offset + 0, data->priv->order,
@@ -418,7 +506,7 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 					EXIF_FORMAT_LONG);
 			exif_set_long  (*d + 6 + offset + 4, data->priv->order,
 					1);
-			exif_set_long  (*d + 6 + offset + 8, data->priv->order, 
+			exif_set_long  (*d + 6 + offset + 8, data->priv->order,
 					data->size);
 			offset += 12;
 		}
@@ -444,62 +532,56 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 }
 
 void
-exif_data_load_data (ExifData *data, const unsigned char *d, unsigned int size)
+exif_data_load_data (ExifData *data, const unsigned char *d_orig,
+		     unsigned int ds_orig)
 {
-	unsigned int l, len = size;
+	unsigned int l;
 	ExifLong offset;
 	ExifShort n;
+	ExifEntry *e, *em;
+	const unsigned char *d = d_orig;
+	unsigned int ds = ds_orig, len;
 
-	if (!data)
-		return;
-	if (!d || !size)
-		return;
+	if (!data || !data->priv || !d || !ds) return;
 
-#ifdef DEBUG
-	printf ("Parsing %i byte(s) EXIF data...\n", size);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Parsing %i byte(s) EXIF data...\n", ds);
 
 	/*
 	 * It can be that the data starts with the EXIF header. If it does
 	 * not, search the EXIF marker.
 	 */
-	if (size < 6) {
-#ifdef DEBUG
-		printf ("Size too small.\n");
-#endif
+	if (ds < 6) {
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		          "Size too small.");
 		return;
 	}
 	if (!memcmp (d, ExifHeader, 6)) {
-#ifdef DEBUG
-		printf ("Found EXIF header.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "Found EXIF header.");
 	} else {
-#ifdef DEBUG
-		printf ("Data begins with 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x "
-			"0x%x...\n", d[0], d[1], d[2], d[3], d[4], d[5], d[6]);
-#endif
 		while (1) {
-			while ((d[0] == 0xff) && size) {
+			while ((d[0] == 0xff) && ds) {
 				d++;
-				size--;
+				ds--;
 			}
 
 			/* JPEG_MARKER_SOI */
 			if (d[0] == JPEG_MARKER_SOI) {
 				d++;
-				size--;
+				ds--;
 				continue;
 			}
 
 			/* JPEG_MARKER_APP0 */
 			if (d[0] == JPEG_MARKER_APP0) {
 				d++;
-				size--;
+				ds--;
 				l = (d[0] << 8) | d[1];
-				if (l > size)
+				if (l > ds)
 					return;
 				d += l;
-				size -= l;
+				ds -= l;
 				continue;
 			}
 
@@ -508,50 +590,45 @@ exif_data_load_data (ExifData *data, const unsigned char *d, unsigned int size)
 				break;
 
 			/* Unknown marker or data. Give up. */
-#ifdef DEBUG
-			printf ("EXIF marker not found.\n");
-#endif
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG,
+				  "ExifData", "EXIF marker not found.");
 			return;
 		}
 		d++;
-		size--;
-		if (size < 2) {
-#ifdef DEBUG
-			printf ("Size too small.\n");
-#endif
+		ds--;
+		if (ds < 2) {
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG,
+				  "ExifData", "Size too small.");
 			return;
 		}
 		len = (d[0] << 8) | d[1];
-#ifdef DEBUG
-		printf ("We have to deal with %i byte(s) of EXIF data.\n", len);
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "We have to deal with %i byte(s) of EXIF data.",
+			  len);
 		d += 2;
-		size -= 2;
+		ds -= 2;
 	}
 
 	/*
 	 * Verify the exif header
 	 * (offset 2, length 6).
 	 */
-	if (size < 6) {
-#ifdef DEBUG
-		printf ("Size too small.\n");
-#endif
+	if (ds < 6) {
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "Size too small.");
 		return;
 	}
 	if (memcmp (d, ExifHeader, 6)) {
-#ifdef DEBUG
-		printf ("EXIF header not found.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "EXIF header not found.");
 		return;
 	}
 
-#ifdef DEBUG
-	printf ("Found EXIF header.\n");
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Found EXIF header.");
 
 	/* Byte order (offset 6, length 2) */
-	if (size < 12)
+	if (ds < 12)
 		return;
 	if (!memcmp (d + 6, "II", 2))
 		data->priv->order = EXIF_BYTE_ORDER_INTEL;
@@ -566,51 +643,93 @@ exif_data_load_data (ExifData *data, const unsigned char *d, unsigned int size)
 
 	/* IFD 0 offset */
 	offset = exif_get_long (d + 10, data->priv->order);
-#ifdef DEBUG
-	printf ("IFD 0 at %i.\n", (int) offset);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData", 
+		  "IFD 0 at %i.", (int) offset);
 
 	/* Parse the actual exif data (offset 14) */
 	exif_data_load_data_content (data, data->ifd[EXIF_IFD_0], d + 6,
-				     size - 6, offset);
+				     ds - 6, offset);
 
 	/* IFD 1 offset */
 	n = exif_get_short (d + 6 + offset, data->priv->order);
 	offset = exif_get_long (d + 6 + offset + 2 + 12 * n, data->priv->order);
 	if (offset) {
-#ifdef DEBUG
-		printf ("IFD 1 at %i.\n", (int) offset);
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "IFD 1 at %i.", (int) offset);
 
 		/* Sanity check. */
-		if (offset > size - 6) {
-#ifdef DEBUG
-			printf ("Bogus offset!\n");
-#endif
+		if (offset > ds - 6) {
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, 
+				  "ExifData", "Bogus offset!");
 			return;
 		}
 
 		exif_data_load_data_content (data, data->ifd[EXIF_IFD_1], d + 6,
-					     size - 6, offset);
+					     ds - 6, offset);
+	}
+
+	/*
+	 * If we got an EXIF_TAG_MAKER_NOTE, try to interpret it. Some
+	 * cameras use pointers in the maker note tag that point to the
+	 * space between IFDs. Here is the only place where we have access
+	 * to that data.
+	 */
+	e = exif_data_get_entry (data, EXIF_TAG_MAKER_NOTE);
+	if (e) {
+
+	    /* Olympus & Nikon */
+	    if ((e->size >= 5) && (!memcmp (e->data, "OLYMP", 5) ||
+				   !memcmp (e->data, "Nikon", 5))) {
+			data->priv->md = exif_mnote_data_olympus_new (data->priv->mem);
+	    } else {
+			char value[7];
+			em = exif_data_get_entry (data, EXIF_TAG_MAKE);
+			/* Pentax & some variant of Nikon */
+			if ((e->size >= 2) && (e->data[0] == 0x00)
+				    && (e->data[1] == 0x1b)) {
+				if (em && !strncasecmp (exif_entry_get_value (em, value, sizeof(value)), "Nikon", 5)) {
+					data->priv->md = exif_mnote_data_olympus_new (data->priv->mem);
+				} else {
+					data->priv->md = exif_mnote_data_pentax_new (data->priv->mem);
+				}
+
+			} else {
+				/* Canon */
+				if (em) {
+		
+					if (!strcmp (exif_entry_get_value (em, value, sizeof(value)), "Canon"))
+						data->priv->md = exif_mnote_data_canon_new (data->priv->mem);
+				}
+			}
+		}
+
+	    /* 
+	     * If we are able to interpret the maker note, do so.
+	     */
+	    if (data->priv->md) {
+		    exif_mnote_data_log (data->priv->md, data->priv->log);
+		    exif_mnote_data_set_byte_order (data->priv->md,
+				    		    data->priv->order);
+		    exif_mnote_data_set_offset (data->priv->md,
+				    		data->priv->offset_mnote);
+		    exif_mnote_data_load (data->priv->md, d, ds);
+	    }
 	}
 }
 
 void
 exif_data_save_data (ExifData *data, unsigned char **d, unsigned int *ds)
 {
-	if (!data)
-		return;
-	if (!d || !ds)
+	if (!data || !d || !ds)
 		return;
 
 	/* Header */
-	*ds = 6;
-	*d = malloc (sizeof (char) * *ds);
+	*ds = 14;
+	*d = exif_data_alloc (data, *ds);
+	if (!*d) return;
 	memcpy (*d, ExifHeader, 6);
 
 	/* Order (offset 6) */
-	*ds += 2;
-	*d = realloc (*d, sizeof (char) * *ds);
 	if (data->priv->order == EXIF_BYTE_ORDER_INTEL) {
 		memcpy (*d + 6, "II", 2);
 	} else {
@@ -618,53 +737,33 @@ exif_data_save_data (ExifData *data, unsigned char **d, unsigned int *ds)
 	}
 
 	/* Fixed value (2 bytes, offset 8) */
-	*ds += 2;
-	*d = realloc (*d, sizeof (char) * *ds);
 	exif_set_short (*d + 8, data->priv->order, 0x002a);
 
 	/*
 	 * IFD 0 offset (4 bytes, offset 10).
 	 * We will start 8 bytes after the
-	 * EXIF header (2 bytes for order, another 2 for the test, and 
+	 * EXIF header (2 bytes for order, another 2 for the test, and
 	 * 4 bytes for the IFD 0 offset make 8 bytes together).
 	 */
-	*ds += 4;
-	*d = realloc (*d, sizeof (char) * *ds);
 	exif_set_long (*d + 10, data->priv->order, 8);
 
 	/* Now save IFD 0. IFD 1 will be saved automatically. */
-#ifdef DEBUG
-	printf ("Saving IFDs...\n");
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saving IFDs...");
 	exif_data_save_data_content (data, data->ifd[EXIF_IFD_0], d, ds,
 				     *ds - 6);
-
-#ifdef DEBUG
-	printf ("Saved %i byte(s) EXIF data.\n", *ds);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saved %i byte(s) EXIF data.", *ds);
 }
 
 ExifData *
 exif_data_new_from_file (const char *path)
 {
-	FILE *f;
-	int size;
 	ExifData *edata;
 	ExifLoader *loader;
-	unsigned char data[1024];
-
-	f = fopen (path, "rb");
-	if (!f)
-		return (NULL);
 
 	loader = exif_loader_new ();
-	while (1) {
-		size = fread (data, 1, 1024, f);
-		if (size <= 0) break;
-		if (!exif_loader_write (loader, data, size)) break;
-	}
-	fclose (f);
-
+	exif_loader_write_file (loader, path);
 	edata = exif_loader_get_data (loader);
 	exif_loader_unref (loader);
 
@@ -696,8 +795,7 @@ exif_data_free (ExifData *data)
 {
 	unsigned int i;
 
-	if (!data)
-		return;
+	if (!data) return;
 
 	for (i = 0; i < EXIF_IFD_COUNT; i++) {
 		if (data->ifd[i]) {
@@ -705,15 +803,21 @@ exif_data_free (ExifData *data)
 			data->ifd[i] = NULL;
 		}
 	}
-	if (data->data) {
-		free (data->data);
-		data->data = NULL;
-	}
+
 	if (data->priv) {
-		free (data->priv);
-		data->priv = NULL;
+		ExifMem *mem = data->priv->mem;
+		if (data->data) {
+			exif_mem_free (data->priv->mem, data->data);
+			data->data = NULL;
+		}
+		if (data->priv->md) {
+			exif_mnote_data_unref (data->priv->md);
+			data->priv->md = NULL;
+		}
+		exif_mem_free (mem, data->priv);
+		exif_mem_free (mem, data);
+		exif_mem_unref (mem);
 	}
-	free (data);
 }
 
 void
@@ -790,7 +894,7 @@ entry_set_byte_order (ExifEntry *e, void *data)
 			s = exif_get_short (e->data +
 				(i * exif_format_get_size (e->format)),
 				d->old);
-			exif_set_short (e->data + 
+			exif_set_short (e->data +
 				(i * exif_format_get_size (e->format)),
 				d->new, s);
 		}
@@ -837,6 +941,7 @@ entry_set_byte_order (ExifEntry *e, void *data)
 		break;
 	case EXIF_FORMAT_UNDEFINED:
 	case EXIF_FORMAT_BYTE:
+	case EXIF_FORMAT_SBYTE:
 	case EXIF_FORMAT_ASCII:
 	default:
 		/* Nothing here. */
@@ -862,4 +967,24 @@ exif_data_set_byte_order (ExifData *data, ExifByteOrder order)
 	d.new = order;
 	exif_data_foreach_content (data, content_set_byte_order, &d);
 	data->priv->order = order;
+	if (data->priv->md)
+		exif_mnote_data_set_byte_order (data->priv->md, order);
+}
+
+void
+exif_data_log (ExifData *data, ExifLog *log)
+{
+	if (!data || !data->priv) return;
+	exif_log_unref (data->priv->log);
+	data->priv->log = log;
+	exif_log_ref (log);
+}
+
+/* Used internally within libexif */
+ExifLog *exif_data_get_log (ExifData *);
+ExifLog *
+exif_data_get_log (ExifData *data)
+{
+	if (!data || !data->priv) return NULL;
+	return data->priv->log;
 }
